/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001, 2002 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *	
 * "$Id:"
 */

#ifndef TVMET_MATRIX_FUNCTION_H
#define TVMET_MATRIX_FUNCTIONS_H 

#include <tvmet/Extremum.h>

namespace tvmet {

/* forwards */
template<class T, std::size_t Sz> class Vector;

/************************************************************************
 * Other Matrix Meta
 ***********************************************************************/

/**
 * \class MetaMM MatrixFunctions.h "tvmet/MatrixFunctions.h"
 * \brief Meta class for matrix-matrix operations. 
 * \note The rows of matrix 2 have to be equal to cols of matrix 1.
 */
template<std::size_t Rows1, std::size_t Cols1, 
	 std::size_t Cols2, 
	 std::size_t RowStride1, std::size_t ColStride1, 
	 std::size_t RowStride2, std::size_t ColStride2, 
	 std::size_t K>
class MetaMM 
{
private:
  MetaMM();
  MetaMM& operator=(const MetaMM<Rows1, Cols1, Cols2, RowStride1, ColStride1, RowStride2, ColStride2, K>&);

private:
  enum { 
    doIt = (K != Cols1 - 1) 		/**< recursive counter */
  };

public:
  template<class T1, class T2>
  static inline 
  typename PromoteTraits<T1, T2>::value_type
  product(const T1* restrict lhs, const T2* restrict rhs, std::size_t i, std::size_t j) {
    return lhs[i * RowStride1 + K * ColStride1] * rhs[K * RowStride2 + j * ColStride2]
      + MetaMM<Rows1 * doIt, Cols1 * doIt,
               Cols2 * doIt, RowStride1 * doIt, ColStride1 * doIt,
               RowStride2 * doIt, ColStride2 * doIt, (K+1) * doIt>::product(lhs, rhs, i, j);
  }
};

/*
 * MetaMM Specialized for recursion
 */
template<>
class MetaMM<0,0,0,0,0,0,0,0> {
  MetaMM();
  MetaMM& operator=(const MetaMM<0,0,0,0,0,0,0,0>&);

public:
  static inline 
  XprNull product(const void*, const void*, std::size_t, std::size_t) {
    return XprNull(); 
  }
};


/**
 * \class MetaMV MatrixFunctions.h "tvmet/MatrixFunctions.h"
 * \brief Meta class for matrix-vector operations
 */
template<std::size_t Rows, std::size_t Cols, std::size_t RowStride, std::size_t ColStride,
	 std::size_t VecStride, std::size_t J>
class MetaMV 
{
  MetaMV();
  MetaMV& operator=(const MetaMV<Rows, Cols, RowStride, ColStride, VecStride, J>&);

private:
  enum {
    doIt = J < (Cols-1)  		/**< recursive counter */
  };
   
public:
  template<class T1, class T2> 
  static inline 
  typename PromoteTraits<T1, T2>::value_type
  product(const T1* restrict m, const T2* restrict v, std::size_t i) {
    return m[i * RowStride + J * ColStride] * v[J * VecStride]
      + MetaMV<Rows * doIt, Cols * doIt,
               RowStride * doIt, ColStride * doIt, 
               VecStride * doIt, (J+1)*doIt>::product(m, v, i);
  }
};

/*
 * MetaMV Specialized for recursion
 */
template<>
class MetaMV<0,0,0,0,0,0> 
{
  MetaMV();
  MetaMV& operator=(const MetaMV<0,0,0,0,0,0>&);

public:
  static inline 
  XprNull product(const void*, const void*, std::size_t) {
    return XprNull(); 
  }
};


/************************************************************************
 * global binary matrix algebra functions
 ***********************************************************************/

/**
 * \fn product(const Matrix<T1, Rows1, Cols1>& lhs, const Matrix<T2, Cols1, Cols2>& rhs) 
 * \brief Function for the matrix-matrix-product. 
 * \ingroup _function
 * \note The rows2 has to be equal to cols1.
 */
template<class T1, std::size_t Rows1, std::size_t Cols1,
	 class T2, std::size_t Cols2>
inline
XprMatrix<
  XprMMProduct<
    T1, Rows1, Cols1, 
    T2, Cols2, Cols1,
    1, Cols2, 1
  >,
  Rows1, Cols2
>
product(const Matrix<T1, Rows1, Cols1>& lhs, const Matrix<T2, Cols1, Cols2>& rhs) {
  typedef XprMMProduct<
    T1, Rows1, Cols1, 
    T2, Cols2, Cols1, 
    1, Cols2, 1
  >							ExprType;
  return XprMatrix<ExprType, Rows1, Cols2>(ExprType(lhs.data(), rhs.data()));
}

/**
 * \fn product(const Matrix<T1, Rows, Cols>& lhs,  const Vector<T2, Cols>& v)
 * \brief Function for the matrix-vector-product
 * \ingroup _function
 */
template<class T1, std::size_t Rows, std::size_t Cols, class T2>
inline
XprVector<
  XprMVProduct<T1, Rows, Cols, Cols, 1, T2, 1>,
  Rows
>
product(const Matrix<T1, Rows, Cols>& m, const Vector<T2, Cols>& v) {
  typedef XprMVProduct<T1, Rows, Cols, Cols, 1, T2, 1> 	ExprType;
  return XprVector<ExprType, Rows>(ExprType(m.data(), v.data()));
}

#if 0 // doesn't work without temporaries yet
/*
 * \fn product(const XprMatrix<E, Rows, Cols>& lhs, const Vector<T, Cols>& v)
 * \brief Compute the product of an XprMatrix with a Vector.
 * \ingroup _function
 */
// XXX shit temporaries
template<class E, std::size_t Rows, std::size_t Cols, class T>
inline
XprVector<
  XprMVProduct<typename E::value_type, Rows, Cols, Cols, 1, T, 1>,
  Rows
>
product(XprMatrix<E, Rows, Cols> lhs, const Vector<T, Cols>& rhs) {
  Matrix<T, Rows, Cols> m;
  m = lhs;
  typedef XprMVProduct<typename E::value_type, Rows, Cols, Cols, 1, T, 1> ExprType;
  return XprVector<ExprType, Rows>(ExprType(m.data(), rhs.data()));
}
#endif

/**
 * \fn productTransposed(const Matrix<T1, Rows, Cols>& matrix, const Vector<T2, Rows>& vector) 
 * \brief Function for the transpose(matrix)-vector-product
 * \ingroup _function
 */
template<class T1, class T2, std::size_t Rows, std::size_t Cols>
inline 
XprVector<
  XprMVProduct<T1, Cols, Rows, 1, Cols, T2, 1>,
  Rows
>
productTransposed(const Matrix<T1, Rows, Cols>& matrix, const Vector<T2, Rows>& vector) {
  typedef XprMVProduct<T1, Cols, Rows, 1, Cols, T2, 1> ExprType;
  return XprVector<ExprType, Rows>(ExprType(matrix.data(), vector.data()));
}

/**
 * \fn transpose(const Matrix<T, Rows, Cols>& rhs)
 * \brief Transpose the matrix
 * \ingroup _function
 */
template<class T, std::size_t Rows, std::size_t Cols>
inline
XprMatrix<
  XprMatrixTranspose<
    MatrixConstReference<T, Rows, Cols> 
  >,
  Cols, Rows
>
transpose(const Matrix<T, Rows, Cols>& rhs) {
  typedef XprMatrixTranspose<
    MatrixConstReference<T, Rows, Cols>
  >							ExprType;
  return XprMatrix<ExprType, Cols, Rows>(ExprType(rhs.cref()));
}


/***********************************************************************
 * global binary math functions
 ***********************************************************************/

/*
 * FUNCTION(Matrix<T1, Rows, Cols>, Matrix<T2, Rows, Cols>)
 */
#define DECLARE_BINARY_FUNCTION(FUNC, FUNCTIONAL)                   	\
template<class T1, class T2, std::size_t Rows, std::size_t Cols>	\
inline                 			      			      	\
XprMatrix<                   						\
  XprBinOp<                   						\
    FUNCTIONAL<T1, T2>,                   				\
    MatrixConstReference<T1, Rows, Cols>,      				\
    MatrixConstReference<T2, Rows, Cols>      				\
  >,                   			              			\
  Rows, Cols				      			      	\
>                   			              			\
FUNC(const Matrix<T1, Rows, Cols>& lhs,           			\
     const Matrix<T2, Cols, Cols>& rhs) {           			\
  typedef XprBinOp <                    				\
    FUNCTIONAL<T1, T2>,  	                 			\
    MatrixConstReference<T1, Rows, Cols>,      				\
    MatrixConstReference<T2, Rows, Cols>       				\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(lhs.cref(), rhs.cref())); \
}

DECLARE_BINARY_FUNCTION(atan2, fcnl_atan2)
DECLARE_BINARY_FUNCTION(drem, fcnl_drem)
DECLARE_BINARY_FUNCTION(fmod, fcnl_fmod)
DECLARE_BINARY_FUNCTION(hypot, fcnl_hypot)
DECLARE_BINARY_FUNCTION(jn, fcnl_jn)
DECLARE_BINARY_FUNCTION(yn, fcnl_yn)
DECLARE_BINARY_FUNCTION(pow, fcnl_pow)
#undef DECLARE_BINARY_FUNCTION

/*
 * Matrix<T, Rows, Cols> and  POD's
 */
#define DECLARE_BINARY_FUNCTION(FUNC, FUNCTIONAL, TP)                 	\
template<class T, std::size_t Rows, std::size_t Cols> 			\
inline                 			      			      	\
XprMatrix<                						\
  XprBinOp<          							\
    FUNCTIONAL<T, TP >,          					\
    MatrixConstReference<T, Rows, Cols>,				\
    XprLiteral< TP >          						\
  >,          								\
  Rows, Cols				      			      	\
>          								\
FUNC(const Matrix<T, Rows, Cols>& lhs, TP rhs) {			\
  typedef XprBinOp<          						\
    FUNCTIONAL<T, TP >,          					\
    MatrixConstReference<T, Rows, Cols>,				\
    XprLiteral< TP >							\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(lhs.cref(), XprLiteral< TP >(rhs))); \
}          								\
          								\
template<class T, std::size_t Rows, std::size_t Cols>			\
inline                 			      			      	\
XprMatrix<          							\
  XprBinOp<          							\
    FUNCTIONAL< TP, T>,          					\
    XprLiteral< TP >,          						\
    MatrixConstReference<T, Rows, Cols>					\
  >,          								\
  Rows, Cols				      			      	\
>          								\
FUNC(TP lhs, const Matrix<T, Rows, Cols>& rhs) {         		\
  typedef XprBinOp<         						\
    FUNCTIONAL< TP, T>,         					\
    XprLiteral< TP >,         						\
    MatrixConstReference<T, Rows, Cols>					\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(XprLiteral< TP >(lhs), rhs.cref())); \
}

DECLARE_BINARY_FUNCTION(atan2, fcnl_atan2, int)
DECLARE_BINARY_FUNCTION(drem, fcnl_drem, int)
DECLARE_BINARY_FUNCTION(fmod, fcnl_fmod, int)
DECLARE_BINARY_FUNCTION(hypot, fcnl_hypot, int)
DECLARE_BINARY_FUNCTION(jn, fcnl_jn, int)
DECLARE_BINARY_FUNCTION(yn, fcnl_yn, int)
DECLARE_BINARY_FUNCTION(pow, fcnl_pow, int)

DECLARE_BINARY_FUNCTION(atan2, fcnl_atan2, float)
DECLARE_BINARY_FUNCTION(drem, fcnl_drem, float)
DECLARE_BINARY_FUNCTION(fmod, fcnl_fmod, float)
DECLARE_BINARY_FUNCTION(hypot, fcnl_hypot, float)
DECLARE_BINARY_FUNCTION(jn, fcnl_jn, float)
DECLARE_BINARY_FUNCTION(yn, fcnl_yn, float)
DECLARE_BINARY_FUNCTION(pow, fcnl_pow, float)

DECLARE_BINARY_FUNCTION(atan2, fcnl_atan2, double)
DECLARE_BINARY_FUNCTION(drem, fcnl_drem, double)
DECLARE_BINARY_FUNCTION(fmod, fcnl_fmod, double)
DECLARE_BINARY_FUNCTION(hypot, fcnl_hypot, double)
DECLARE_BINARY_FUNCTION(jn, fcnl_jn, double)
DECLARE_BINARY_FUNCTION(yn, fcnl_yn, double)
DECLARE_BINARY_FUNCTION(pow, fcnl_pow, double)

#ifdef HAVE_LONG_DOUBLE
DECLARE_BINARY_FUNCTION(atan2, fcnl_atan2, long double)
DECLARE_BINARY_FUNCTION(drem, fcnl_drem, long double)
DECLARE_BINARY_FUNCTION(fmod, fcnl_fmod, long double)
DECLARE_BINARY_FUNCTION(hypot, fcnl_hypot, long double)
DECLARE_BINARY_FUNCTION(jn, fcnl_jn, long double)
DECLARE_BINARY_FUNCTION(yn, fcnl_yn, long double)
DECLARE_BINARY_FUNCTION(pow, fcnl_pow, long double)
#endif // HAVE_LONG_DOUBLE

#undef DECLARE_BINARY_FUNCTION

#ifdef HAVE_COMPLEX
/*
 * Matrix<T, Rows, Cols> and  std::complex<>
 */
#define DECLARE_BINARY_FUNCTION(FUNC, FUNCTIONAL)                 	\
template<class T, std::size_t Rows, std::size_t Cols> 			\
inline                 			      			      	\
XprMatrix<                						\
  XprBinOp<          							\
    FUNCTIONAL<T, std::complex<T> >,          				\
    MatrixConstReference<T, Rows, Cols>,				\
    XprLiteral< std::complex<T> >          				\
  >,          								\
  Rows, Cols				      			      	\
>          								\
FUNC(const Matrix<T, Rows, Cols>& lhs, const std::complex<T>& rhs) {	\
  typedef XprBinOp<          						\
    FUNCTIONAL<T, std::complex<T> >,          				\
    MatrixConstReference<T, Rows, Cols>,				\
    XprLiteral< std::complex<T> >					\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(lhs.cref(), XprLiteral< std::complex<T> >(rhs))); \
}          								\
          								\
template<class T, std::size_t Rows, std::size_t Cols>			\
inline                 			      			      	\
XprMatrix<          							\
  XprBinOp<          							\
    FUNCTIONAL< std::complex<T>, T>,          				\
    XprLiteral< std::complex<T> >,          				\
    MatrixConstReference<T, Rows, Cols>					\
  >,          								\
  Rows, Cols				      			      	\
>          								\
FUNC(const std::complex<T>& lhs, const Matrix<T, Rows, Cols>& rhs) {	\
  typedef XprBinOp<         						\
    FUNCTIONAL< std::complex<T>, T>,         				\
    XprLiteral< std::complex<T> >,         				\
    MatrixConstReference<T, Rows, Cols>					\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(XprLiteral< std::complex<T> >(lhs), rhs.cref())); \
}

DECLARE_BINARY_FUNCTION(atan2, fcnl_atan2)
DECLARE_BINARY_FUNCTION(drem, fcnl_drem)
DECLARE_BINARY_FUNCTION(fmod, fcnl_fmod)
DECLARE_BINARY_FUNCTION(hypot, fcnl_hypot)
DECLARE_BINARY_FUNCTION(jn, fcnl_jn)
DECLARE_BINARY_FUNCTION(yn, fcnl_yn)
DECLARE_BINARY_FUNCTION(pow, fcnl_pow)
#undef DECLARE_BINARY_FUNCTION

#endif // HAVE_COMPLEX

/*
 * Matrix<T, Rows, Cols> and  XprMatrix<E, Rows, Cols>
 */
#define DECLARE_BINARY_FUNCTION(FUNC, FUNCTIONAL)                   	\
template<class E, class T, std::size_t Rows, std::size_t Cols> 		\
inline                 			      			      	\
XprMatrix<                    	      					\
  XprBinOp<                     	      				\
    FUNCTIONAL<typename E::value_type, T>,                    	      	\
    MatrixConstReference<T, Rows, Cols>,       	      			\
    XprMatrix<E, Rows, Cols>						\
  >,                     	      					\
  Rows, Cols				      			      	\
>                      	      						\
FUNC(XprMatrix<E, Rows, Cols> lhs, const Matrix<T, Rows, Cols>& rhs) {	\
  typedef XprBinOp<                     	      			\
    FUNCTIONAL<typename E::value_type, T>,                    	      	\
    XprMatrix<E, Rows, Cols>,          	      				\
    MatrixConstReference<T, Rows, Cols>        	      			\
  > 							 ExprType; 	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(lhs, rhs.cref()));	\
}                    	      						\
                    	      						\
template<class E, class T, std::size_t Rows, std::size_t Cols> 		\
inline                 			      			      	\
XprMatrix<                    	      					\
  XprBinOp<                     	      				\
    FUNCTIONAL<T, typename E::value_type>,				\
    MatrixConstReference<T, Rows, Cols>,       	      			\
    XprMatrix<E, Rows, Cols>						\
  >,                     	      					\
  Rows, Cols				      			      	\
>                      	      						\
FUNC(const Matrix<T, Rows, Cols>& lhs, XprMatrix<E, Rows, Cols> rhs) {	\
  typedef XprBinOp<                     	      			\
    FUNCTIONAL<T, typename E::value_type>,                    	      	\
    MatrixConstReference<T, Rows, Cols>,       	      			\
    XprMatrix<E, Rows, Cols>						\
  > 						 	ExprType; 	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(lhs.cref(), rhs));	\
}

DECLARE_BINARY_FUNCTION(atan2, fcnl_atan2)
DECLARE_BINARY_FUNCTION(drem, fcnl_drem)
DECLARE_BINARY_FUNCTION(fmod, fcnl_fmod)
DECLARE_BINARY_FUNCTION(hypot, fcnl_hypot)
DECLARE_BINARY_FUNCTION(jn, fcnl_jn)
DECLARE_BINARY_FUNCTION(yn, fcnl_yn)
DECLARE_BINARY_FUNCTION(pow, fcnl_pow)
#undef DECLARE_BINARY_FUNCTION

/***********************************************************************
 * global unary math functions
 ***********************************************************************/

/*
 * Functions on  Matrix<T, Rows, Cols>
 */
#define DECLARE_UNARY_FUNCTION(FUNC, FUNCTIONAL)			\
template<class T, std::size_t Rows, std::size_t Cols>			\
inline									\
XprMatrix<								\
  XprUnOp<								\
    FUNCTIONAL<T>, 							\
    MatrixConstReference<T, Rows, Cols>					\
  >, 									\
  Rows, Cols				      			      	\
>									\
FUNC(const Matrix<T, Rows, Cols>& rhs) {				\
  typedef XprUnOp<							\
    FUNCTIONAL<T>, 							\
    MatrixConstReference<T, Rows, Cols>					\
  > 							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(rhs.cref()));		\
}

DECLARE_UNARY_FUNCTION(abs, fcnl_abs)
DECLARE_UNARY_FUNCTION(cbrt, fcnl_cbrt)
DECLARE_UNARY_FUNCTION(ceil, fcnl_ceil)
DECLARE_UNARY_FUNCTION(floor, fcnl_floor)
DECLARE_UNARY_FUNCTION(rint, fcnl_rint)
DECLARE_UNARY_FUNCTION(sin, fcnl_sin)
DECLARE_UNARY_FUNCTION(cos, fcnl_cos)
DECLARE_UNARY_FUNCTION(tan, fcnl_tan)
DECLARE_UNARY_FUNCTION(sinh, fcnl_sinh)
DECLARE_UNARY_FUNCTION(cosh, fcnl_cosh)
DECLARE_UNARY_FUNCTION(tanh, fcnl_tanh)
DECLARE_UNARY_FUNCTION(asin, fcnl_asin)
DECLARE_UNARY_FUNCTION(acos, fcnl_acos)
DECLARE_UNARY_FUNCTION(atan, fcnl_atan)
DECLARE_UNARY_FUNCTION(exp, fcnl_exp)
DECLARE_UNARY_FUNCTION(log, fcnl_log)
DECLARE_UNARY_FUNCTION(log10, fcnl_log10)
DECLARE_UNARY_FUNCTION(sqrt, fcnl_sqrt)

#ifdef HAVE_IEEE_MATH
DECLARE_UNARY_FUNCTION(asinh, fcnl_asinh)
DECLARE_UNARY_FUNCTION(acosh, fcnl_acosh)
DECLARE_UNARY_FUNCTION(atanh, fcnl_atanh)
DECLARE_UNARY_FUNCTION(expm1, fcnl_expm1)
DECLARE_UNARY_FUNCTION(log1p, fcnl_log1p)
DECLARE_UNARY_FUNCTION(erf, fcnl_erf)
DECLARE_UNARY_FUNCTION(erfc, fcnl_erfc)
DECLARE_UNARY_FUNCTION(j0, fcnl_j0)
DECLARE_UNARY_FUNCTION(j1, fcnl_j1)
DECLARE_UNARY_FUNCTION(y0, fcnl_y0)
DECLARE_UNARY_FUNCTION(y1, fcnl_y1)
DECLARE_UNARY_FUNCTION(lgamma, fcnl_lgamma)
// XXX default return is only an int!
DECLARE_UNARY_FUNCTION(isnan, fcnl_isnan)
DECLARE_UNARY_FUNCTION(isinf, fcnl_isinf)
DECLARE_UNARY_FUNCTION(finite, fcnl_finite)
#endif // HAVE_IEEE_MATH

#undef DECLARE_UNARY_FUNCTION

/**
 * \fn max(const XprMatrix<E, Rows, Cols>& e)
 * \brief Find the maximum of a matrix expression
 * \ingroup _function
 */
template<class E, std::size_t Rows, std::size_t Cols>
inline
Extremum<typename E::value_type, std::size_t, matrix_tag>
max(const XprMatrix<E, Rows, Cols>& e) {
  typedef typename E::value_type value_type;

  value_type max( e(0, 0) );
  std::size_t row(0), col(0);

  // this loop is faster than meta templates!
  for(std::size_t i = 1; i < Rows; ++i) {
    for(std::size_t j = 1; j < Cols; ++j) {
      if(e(i, j) > max) {
	max = e(i, j);
	row = i;
	col = j;
      }
    }
  }

  return Extremum<value_type, std::size_t, matrix_tag>(max, row, col);
}

/**
 * \fn max(const Matrix<T, Rows, Cols>& m)
 * \brief Find the maximum of a matrix
 * \ingroup _function
 */
template<class T, std::size_t Rows, std::size_t Cols>
inline
Extremum<T, std::size_t, matrix_tag>
max(const Matrix<T, Rows, Cols>& m) { return max(m.asXpr()); }

/**
 * \fn min(const XprMatrix<E, Rows, Cols>& e)
 * \brief Find the minimum of a matrix expression
 * \ingroup _function
 */
template<class E, std::size_t Rows, std::size_t Cols>
inline
Extremum<typename E::value_type, std::size_t, matrix_tag>
min(const XprMatrix<E, Rows, Cols>& e) {
  typedef typename E::value_type value_type;

  value_type min( e(0, 0) );
  std::size_t row(0), col(0);

  // this loop is faster than meta templates!
  for(std::size_t i = 1; i < Rows; ++i) {
    for(std::size_t j = 1; j < Cols; ++j) {
      if(e(i, j) < min) {
	min = e(i, j);
	row = i;
	col = j;
      }
    }
  }

  return Extremum<value_type, std::size_t, matrix_tag>(min, row, col);
}

/**
 * \fn min(const Matrix<T, Rows, Cols>& m)
 * \brief Find the minimum of a matrix
 * \ingroup _function
 */
template<class T, std::size_t Rows, std::size_t Cols>
inline
Extremum<T, std::size_t, matrix_tag>
min(const Matrix<T, Rows, Cols>& m) { return min(m.asXpr()); }

/**
 * \fun row(Matrix<T, Rows, Cols>& m, std::size_t no)
 * \brief Returns a row vector of the given matrix.
 * \ingroup _function
 */
template<class T, std::size_t Rows, std::size_t Cols>
inline
XprVector<
  MatrixRowVectorReference<T, Rows, Cols>,
  Cols
>
row(Matrix<T, Rows, Cols>& m, std::size_t no) {
  typedef MatrixRowVectorReference<T, Rows, Cols>	ExprType;
  return XprVector<ExprType, Cols>(ExprType(m.data(), no));
}

/**
 * \fun col(Matrix<T, Rows, Cols>& m, std::size_t no)
 * \brief Returns a column vector of the given matrix.
 * \ingroup _function
 */
template<class T, std::size_t Rows, std::size_t Cols>
inline
XprVector<
  MatrixColVectorReference<T, Rows, Cols>,
  Rows
>
col(Matrix<T, Rows, Cols>& m, std::size_t no) {
  typedef MatrixColVectorReference<T, Rows, Cols>	ExprType;
  return XprVector<ExprType, Rows>(ExprType(m.data(), no));
}

} // namespace tvmet

#endif // TVMET_MATRIX_FUNCTIONS_H

// Local Variables:
// mode:C++
// End:
